from sqlalchemy import *

from dataProviderBase import *
from utils import *
from analysis import *
from consts import *
from sqlManager import *

import globals
import config

class RunsDataProvider(PagedTableDataProvider):
    RUN_ID_TITLE = "Run ID"
    STREAMERS_TITLE = "Streamers"
    REPACKED_TITLE = "Repacked"
    RECO_TITLE = "Reconstructed"
    REPACK_MERGE_TITLE = "Merged"

    cols = Cols([
        Col(RUN_ID_TITLE, decorator=formRunLink),
        Col("Start", decorator=timeToClock),
        Col("End", decorator=timeToClock),
        Col("Process"),
        Col("Version"),
        Col("Reco version"),
        Col("Global tag"),
        Col("Status", info=
            """
                <i style="color:blue;font-size:10px">Active</i> - Run is ongoing and streamer data is still arriving for it<br />
                <i style="color:blue;font-size:10px">CloseOutSegmentInjector</i> - The Tier-0 is trying to make consistent luminosity sections to be repacked (currently we need 4*n streamer files per Luminosity Section)<br />
                <i style="color:blue;font-size:10px">CloseOutScheduler</i> - There is streamer data for the run that has not been assigned to a repack job yet<br />
                <i style="color:blue;font-size:10px">CloseOutRepack</i> - There is streamer data in a run that has not been repacked yet (create repack jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutRepackMerge</i> - There is RAW data in the run that needs to be merged (create repack merge jobs and wait from them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptReco</i> - There is RAW data in the run that needs to be reconstructed (create prompt reco jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">Complete</i> - We have finished processing this run (there could have been failures along the way but we've exhausted all automatic retries)
            """
        ),
        Col(STREAMERS_TITLE),
        Col(REPACKED_TITLE, info=
            """
                The status bar and percentage displays how many streamer
                files were repacked<br /><br />The number shows how many
                repack files were produced
            """
        ),
        Col(REPACK_MERGE_TITLE),
        Col(RECO_TITLE, info=
            """
                The status bar and percentage displays how many merged
                files, marked for reconstruction were
                actually reconstructed<br /><br />The number shows how many
                merge files were produced
            """
        ),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = 10):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize)
        PagedTableDataProvider._load(self)
        self.title = str(self.dataCount) + " runs"

    def _analysis(self):
        for line in self.data:
            runid = line[self.cols.getId(self.RUN_ID_TITLE)]
            numStreamers = line[self.cols.getId(self.STREAMERS_TITLE)]
            numRepacked = line[self.cols.getId(self.REPACKED_TITLE)]
            numMerge = line[self.cols.getId(self.REPACK_MERGE_TITLE)]
            numReco = line[self.cols.getId(self.RECO_TITLE)]
            # analyse repacked percentage
            repackedP = formatPercentage(numRepacked, getRunRepackedPercentage(runid, numStreamers))
            # analyse merged percentage
            mergedP = formatPercentage(numMerge, getRunRepackMergePercentage(runid, numRepacked))
            # analyse reco percentage
            recoP = formatPercentage(numReco, getRunRecoPercentage(runid, numReco))

            line[self.cols.getId(self.REPACK_MERGE_TITLE)] = mergedP
            line[self.cols.getId(self.REPACKED_TITLE)] = repackedP
            line[self.cols.getId(self.RECO_TITLE)] = recoP

    def _buildCountQuery(self):
        runTable = Table('run', globals.meta, autoload = True, schema = config.DB_SCHEMA)

        self.countQuery = select(
            [func.count(runTable.c.run_id)]
        )

    def _buildMainQuery(self):
        runTable = S_TABLES.RUN
        runStatusTable = S_TABLES.RUN_STATUS
        streamerTable = S_TABLES.STREAMER
        repackedTable = S_TABLES.REPACKED
        reconstructedTable = S_TABLES.RECONSTRUCTED

        if self.asc: order = "ASC"
        else: order = "DESC"

        # TODO:1:improve merge and repacked count queries
        # HACK: query time optimisation
        if self.orderby < self.cols.getId(self.STREAMERS_TITLE):
            self.mainQuery = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.r_version, sub.tag, sub.status,
            (SELECT COUNT(*) FROM """ + streamerTable + """ s WHERE s.run_id = sub.run_id) numStreamers,
            (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = sub.run_id AND re.repacked_id NOT IN (SELECT output_id FROM """ + S_TABLES.REPACKED_MERGE_PARENTAGE + """)) numRepacked,
            (SELECT COUNT(*) FROM """ + repackedTable + """ re WHERE (re.status = """ + str(REPACK_RECONSTRUCTION_QUEUED) + """ OR re.status = """ + str(REPACK_RECONSTRUCTION_DISABLED) + """)  AND re.run_id = sub.run_id),
            (SELECT COUNT(*) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = sub.run_id) numReconstructed
            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.reco_version r_version,
                r.global_tag tag,
                rs.status status
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub"""
        else:
            # slow query
            self.mainQuery = """
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.reco_version,
                r.global_tag,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = r.run_id AND re.repacked_id NOT IN (SELECT output_id FROM """ + S_TABLES.REPACKED_MERGE_PARENTAGE + """)) numRepacked,
                (SELECT COUNT(*) FROM """ + repackedTable + """ re WHERE (re.status = """ + str(REPACK_RECONSTRUCTION_QUEUED) + """ OR re.status = """ + str(REPACK_RECONSTRUCTION_DISABLED) + """)  AND re.run_id = r.run_id),
                (SELECT COUNT(rec.run_id) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = r.run_id) numReconstructed
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order

#        globals.logger.log(self.query)
